/*
    Copyright 2019 Harald Sitter <sitter@kde.org>

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation; either version 3 of
    the License or any later version accepted by the membership of
    KDE e.V. (or its successor approved by the membership of KDE
    e.V.), which shall act as a proxy defined in Section 14 of
    version 3 of the license.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>
*/

#include <QApplication>
#include <QDebug>
#include <QMetaProperty>
#include <QPointer>
#include <QProcess>
#include <QWidget>
#include <QPushButton>
#include <QAbstractListModel>

#include <libgupnp/gupnp-control-point.h>

#define STRING_PROP(name) \
private: \
    Q_PROPERTY(QString name READ name) \
public: \
    QString name() const { return QString::fromUtf8(gupnp_device_info_get_##name(m_info)); } \
private:

class UpnpDeviceInfo : public QObject
{
    Q_OBJECT
    // Only mapping what we need
    STRING_PROP(udn)
    STRING_PROP(friendly_name)
    STRING_PROP(manufacturer)
    STRING_PROP(manufacturer_url)
    STRING_PROP(model_name)
    STRING_PROP(model_number)
    STRING_PROP(model_url)
    STRING_PROP(presentation_url)
    STRING_PROP(serial_number)
public:
    typedef QPointer<UpnpDeviceInfo> Ptr;

    UpnpDeviceInfo(GUPnPDeviceProxy *proxy, QObject *parent = nullptr)
        : QObject(parent)
        , m_proxy(proxy)
        , m_info(GUPNP_DEVICE_INFO(proxy))
    {
        g_object_ref(m_proxy);
    }

    ~UpnpDeviceInfo()
    {
        g_object_unref(m_proxy);
    }

private:
    GUPnPDeviceProxy *m_proxy = nullptr;
    GUPnPDeviceInfo *m_info = nullptr;
};

class UpnpDiscoverer : public QObject
{
    Q_OBJECT
public:
    void discover()
    {
        qDebug() << "STAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAART";
        // https://developer.gnome.org/gupnp/unstable/client-tutorial.html
        GError *error = nullptr;
        auto context = gupnp_context_new(nullptr, "br0", 0, &error);
#warning fixme should just mark defunct
        Q_ASSERT_X(!error, "", error->message);
        auto cp = gupnp_control_point_new(context, "ssdp:all");
        g_signal_connect(cp, "service-proxy-available",
                         G_CALLBACK(+[](GUPnPControlPoint *control_point,
                                    GUPnPServiceProxy *proxy,
                                    gpointer           user_data)
        {
                             auto that = static_cast<UpnpDiscoverer *>(user_data);
                             that->serviceProxyAvailable(control_point, proxy);
                         }), this);
        g_signal_connect(cp, "device-proxy-available",
                         G_CALLBACK(+[](GUPnPControlPoint *control_point,
                                    GUPnPDeviceProxy  *proxy,
                                    gpointer           user_data)
        {
                             auto that = static_cast<UpnpDiscoverer *>(user_data);
                             that->deviceProxyAvailable(control_point, proxy);
                         }), this);
        gssdp_resource_browser_set_active(GSSDP_RESOURCE_BROWSER(cp), TRUE);
    }

public Q_SLOTS:
    void serviceProxyAvailable(GUPnPControlPoint *control_point,
                               GUPnPServiceProxy *proxy)
    {
        qDebug() << "new service";

    }

    void deviceProxyAvailable(GUPnPControlPoint *control_point,
                              GUPnPDeviceProxy *proxy)
    {
        qDebug() << "new device";
        Q_EMIT deviceAvailable(UpnpDeviceInfo::Ptr(new UpnpDeviceInfo(proxy)));
    }

Q_SIGNALS:
    void deviceAvailable(UpnpDeviceInfo::Ptr info);
};

int main(int argc, char **argv)
{
    QApplication app(argc, argv);
    app.setApplicationName("kde-upnpviewer");
    UpnpDiscoverer d;
    d.discover();
    QPushButton b;
    QObject::connect(&b, &QPushButton::clicked, [&](){
        app.quit();
    });
    b.show();
    return app.exec();
}

#include "main.moc"
