/*
    Copyright 2019 Harald Sitter <sitter@kde.org

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation; either version 3 of
    the License or any later version accepted by the membership of
    KDE e.V. (or its successor approved by the membership of KDE
    e.V.), which shall act as a proxy defined in Section 14 of
    version 3 of the license.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "dnssddiscoverer.h"

#include <QUrl>

#include <kio/udsentry.h>
#include <functional>

DNSSDDiscovery::DNSSDDiscovery(KDNSSD::RemoteService::Ptr service)
    : m_service(service)
{
}

QString DNSSDDiscovery::displayType() const
{
    static const QMap<QString, QString> map
    {
        { "_ftp", "FTP" },
        { "_sftp-ssh", "SFTP-SSH" },
        { "_ftps", "Secure FTP" },
        { "_nfs", "NFS" },
        { "_afpovertcp", "afpovertcp" },
        { "_webdav", "WebDAV" },
        { "_webdavs", "Secure WebDAV" },


        { "_workstation", "Workstation" },
        { "_ipp", "Internet Printer (IPP)" },
        { "_printer", "UNIX Printer" },

        { "_ssh", "SSH Remote Terminal" },
        { "_smb", "Microsoft Windows Share"}
    };

    const auto proto = protocol();
    return map.value(proto, proto);
}

QString DNSSDDiscovery::icon() const
{
    static const QMap<QString, QString> map
    {
        { "_ipp", "printer" },
        { "_printer", "printer" },

        { "_ssh", "utilities-terminal" }
    };

    const auto proto = protocol();
    return map.value(proto, "network-server");
}

//QString DNSSDDiscovery::url() const
//{
//    static const QMap<QString, std::function<QString()>> map
//    {
//        { "_ssh", [this] { return "123"; } }
//    };

//    const auto proto = protocol();
//    return map.value(proto, "network-server");
//}

QString DNSSDDiscovery::protocol() const
{
    const auto type = m_service->type();
    return type.split(".").first();
}

void DNSSDDiscovery::toEntry(KIO::UDSEntry &entry)
{
    entry.fastInsert(KIO::UDSEntry::UDS_NAME, m_service->serviceName());
#warning i18n
    entry.fastInsert(KIO::UDSEntry::UDS_DISPLAY_TYPE, displayType());
    entry.fastInsert(KIO::UDSEntry::UDS_ICON_NAME, icon());
    entry.fastInsert(KIO::UDSEntry::UDS_ICON_OVERLAY_NAMES, "/home/me/avahi.png");

    QUrl url;
    url.setScheme("discovery-view");
    url.setHost("dnssd");
    url.setPath("/" + m_service->domain() + "/" + m_service->type() + "/" + m_service->serviceName());
    qDebug() << url;

    entry.fastInsert(KIO::UDSEntry::UDS_URL, url.toString());
}

QHostAddress DNSSDDiscovery::address()
{
    #warning kdnssd needs an address resolver so we can map all services to the same device me thinks
    const auto hostName = m_service->hostName();
    return KDNSSD::ServiceBrowser::resolveHostName(hostName);
}

KDNSSD::RemoteService::Ptr DNSSDDiscovery::service() const
{
    return m_service;
}

DNSSDDiscoverer::DNSSDDiscoverer()
{
    connect(&m_browser, &KDNSSD::ServiceTypeBrowser::serviceTypeAdded,
            this, [=](const QString &type){
        qDebug() << "DNSSD type added:" << type;

        if (m_serviceBrowsers.contains(type)) {
            return; // already exists
        }

        auto browser = new KDNSSD::ServiceBrowser(type);

        connect(browser, &KDNSSD::ServiceBrowser::serviceAdded,
                this, [=](KDNSSD::RemoteService::Ptr service){
            qDebug() << "DNSSD added:"
                             << service->serviceName()
                             << service->type()
                             << service->domain()
                             << service->hostName()
                             << service->port();
            // Manual contains check. We need to use the == of the underlying
            // objects, not the pointers. The same service may have >1
            // RemoteService* instances representing it, so the == impl of
            // RemoteService::Ptr is useless here.
            for (const auto &it : m_services) {
                if (*service == *it) {
                    return;
                }
            }

            connect(service.data(), &KDNSSD::RemoteService::resolved,
                    this, [=] {
                Q_EMIT newDiscovery(Discovery::Ptr(new DNSSDDiscovery(service)));
            });

            // Schedule resolution of hostname. We'll later call resolve
            // which will block until the resolution is done. This basically
            // gives us a head start on discovery.
            service->resolveAsync();
            m_services.append(service);
        });

#warning todo
//        connect(browser, &KDNSSD::ServiceBrowser::finished,
//                this, [=] {
//            qDebug() << "DNSSD finito";
//            m_finished = true;
//            Q_EMIT finished();
//        });

        m_serviceBrowsers[type] = browser;
        browser->startBrowse();
    });
}

void DNSSDDiscoverer::start()
{
    qDebug() << "DNSSD!!!";
    m_browser.startBrowse();
}

void DNSSDDiscoverer::stop()
{
    m_browser.disconnect();
    for (auto service : m_services) {
        service->resolve(); // Blocks until resolution happened. Our signal handle then jumps in.
    }
}

bool DNSSDDiscoverer::isFinished()
{
    return m_finished;
}
