/* This file is part of the KDE project
   Copyright (C) 2003 Joseph Wenninger <jowenn@kde.org>
   Copyright (C) 2008 David Faure <faure@kde.org>

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public License
   along with this library; see the file COPYING.LIB.  If not, write to
   the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
   Boston, MA 02110-1301, USA.
*/

#include <QDebug>
#include <QTimer>
#include <QStandardPaths>

#include <kservicetypetrader.h>
#include <kio/slavebase.h>
#include <kservice.h>
#include <kservicegroup.h>

#include <kio/listjob.h>

#include <time.h>
#include <optional>

#include "upnp.h"
#include "wsdiscoverer.h"
#include "dnssddiscoverer.h"

struct Error
{
    int code = KJob::NoError;
    QString msg;
};

struct Device
{
    explicit Device() = default;
    ~Device() = default;

    void addDiscovery(Discovery::Ptr discovery)
    {
        m_discoveries << discovery;
    }

    QList<Discovery::Ptr> m_discoveries;
};

struct DNSSDDevice
{
    explicit DNSSDDevice(const Discovery::Ptr &discovery)
    {
        addDiscovery(discovery);
    }
    ~DNSSDDevice() = default;

    void addDiscovery(const Discovery::Ptr &discovery)
    {
        const DNSSDDiscovery *d = dynamic_cast<DNSSDDiscovery *>(discovery.data());
        const auto service = d->service();

        const auto name = service->hostName();
        const auto address = KDNSSD::ServiceBrowser::resolveHostName(name);
        if (!m_names.contains(name)) {
            m_names << name;
        }
        if (!m_addresses.contains(address)) {
            m_addresses << address;
        }

        m_discoveries << discovery;
    }

    bool coversDiscovery(const Discovery::Ptr &discovery) const
    {
        const DNSSDDiscovery *d = dynamic_cast<DNSSDDiscovery *>(discovery.data());
        const auto service = d->service();

        const auto name = service->hostName();
        const auto address = KDNSSD::ServiceBrowser::resolveHostName(name);
        return m_names.contains(name) || m_addresses.contains(address);
    }

    QList<Discovery::Ptr> m_discoveries;
    QList<QString> m_names;
    QList<QHostAddress> m_addresses;
};

class Internal
{
public:
    Internal(KIO::SlaveBase *qptr);
    ~Internal() = default;
    std::optional<Error> get(const QUrl& url) Q_REQUIRED_RESULT;
    std::optional<Error> stat(const QUrl& url) Q_REQUIRED_RESULT;
    std::optional<Error> listDir(const QUrl& url) Q_REQUIRED_RESULT;

private:
    KIO::SlaveBase *q = nullptr;
};


Internal::Internal(KIO::SlaveBase *qptr)
    : q(qptr)
{
}


static QList<DNSSDDevice> dnssdDevices;

std::optional<Error> Internal::stat(const QUrl& url)
{
    qDebug() << url;

    for (const auto &device : dnssdDevices) {
        if (!url.path().contains(device.m_names.at(0))) {
            continue;
        }
        KIO::UDSEntry entry;
        entry.fastInsert(KIO::UDSEntry::UDS_FILE_TYPE, S_IFDIR);
        entry.fastInsert(KIO::UDSEntry::UDS_NAME, device.m_names.at(0));
        entry.fastInsert(KIO::UDSEntry::UDS_ICON_NAME, "network-server");
        entry.fastInsert(KIO::UDSEntry::UDS_DISPLAY_TYPE, "DNS-SD Workstation");
        entry.fastInsert(KIO::UDSEntry::UDS_ICON_OVERLAY_NAMES, "/home/me/avahi.png");
        q->statEntry(entry);
        break;
    }

    return std::nullopt;
}

std::optional<Error> Internal::listDir(const QUrl &dirytUrl)
{
    const QUrl url = dirytUrl.adjusted(QUrl::NormalizePathSegments);
    qDebug() << url;

    if (url.path().size() > 1) { // more than /
        for (const auto &device : dnssdDevices) {
            if (!url.path().contains(device.m_names.at(0))) {
                continue;
            }

            KIO::UDSEntry entry;
            for (const auto &discovery : device.m_discoveries) {
                discovery->toEntry(entry);
                q->listEntries({ entry }); // listEntries sends immediately
                entry.clear();
            }

            break;
        }
        return std::nullopt;
    }

    QEventLoop loop;

    UpnpDiscoverer discoverer;
    WSDiscoverer wsdiscoverer;
    DNSSDDiscoverer dnssddiscoverer;

    QTimer quitTimer;
    QObject::connect(&quitTimer, &QTimer::timeout,
                     &loop, &QEventLoop::quit);
    quitTimer.start(5000);

    QList<Device> devices;

    KIO::UDSEntry entry;
    QObject::connect(&discoverer, &UpnpDiscoverer::deviceAvailable,
                     [this, &entry](UpnpDeviceInfo *device) {
        entry.fastInsert(KIO::UDSEntry::UDS_NAME, device->device_type());
#warning i18n
#warning should likely be in device class instead
        QString displayName = device->friendly_name();
        if (displayName.isEmpty()) {
            displayName = "Unknown Device";
        }
        entry.fastInsert(KIO::UDSEntry::UDS_DISPLAY_NAME, displayName);
        entry.fastInsert(KIO::UDSEntry::UDS_ICON_NAME, "network-modem");
//        entry.fastInsert( KIO::UDSEntry::UDS_MIME_TYPE,);
        entry.fastInsert(KIO::UDSEntry::UDS_DISPLAY_TYPE, "Router");
        entry.fastInsert(KIO::UDSEntry::UDS_XML_PROPERTIES, "XML");
        entry.fastInsert(KIO::UDSEntry::UDS_COMMENT, "Fake Comment");
        entry.fastInsert(KIO::UDSEntry::UDS_EXTRA, "Extra Data Blob");
        entry.fastInsert(KIO::UDSEntry::UDS_ICON_OVERLAY_NAMES, "/home/me/Upnp_logo.jpg");

        if (!device->presentation_url().isEmpty()) {
            entry.fastInsert(KIO::UDSEntry::UDS_URL, device->presentation_url());
        } else { // construct a viewer url instead
            // This isn't really ideal. What if you want to view a resource but also
            // interact with it. I think this needs some magic in dolphin or kio
            // to inject viewing capabilities somehow.
            QUrl url;
            url.setScheme("discovery-view");
            url.setHost("upnp");
            url.setPath(QStringLiteral("/") + device->udn());
            entry.fastInsert(KIO::UDSEntry::UDS_URL, url.toString(QUrl::FullyEncoded));
        }

        if (device->friendly_name().contains("urn:")) { // Not really friendly -> hide
            entry.fastInsert(KIO::UDSEntry::UDS_HIDDEN, 1);
        }

        q->listEntries({ entry }); // listEntries sends immediately
        entry.clear();
    });

    QObject::connect(&wsdiscoverer, &WSDiscoverer::newDiscovery,
                     [this, &entry](Discovery::Ptr discovery) {
        qDebug() << "NEW WSDISC";
        discovery->toEntry(entry);
        q->listEntries({ entry }); // listEntries sends immediately
        entry.clear();
    });
    QList<Discovery::Ptr> dnssdDiscoveries;
    QObject::connect(&dnssddiscoverer, &DNSSDDiscoverer::newDiscovery,
                     [this, &entry, &dnssdDiscoveries](Discovery::Ptr discovery) {
        qDebug() << "NEW DNDISC";
        dnssdDiscoveries << discovery;
//        discovery->toEntry(entry);
//        q->listEntries({ entry });  // listEntries sends immediately
//        entry.clear();
    });

    wsdiscoverer.start();
    discoverer.discover();
    dnssddiscoverer.start();

    loop.exec();

    // Iterate the dnssd services and map them into devices

    dnssdDevices.clear();
    for (const auto &discovery : qAsConst(dnssdDiscoveries)) {
        bool covered = false;
        for (auto &device : dnssdDevices) {
            if (device.coversDiscovery(discovery)) {
                covered = true;
                device.addDiscovery(discovery);
                break;
            }
        }
        if (covered) {
            continue;
        }

        dnssdDevices << DNSSDDevice(discovery);
    }

    for (const auto &device : qAsConst(dnssdDevices)) {
        qDebug() << "DNSSDDevice names" << device.m_names;
#warning this isnt technically a dir and screws with property expectations
        entry.fastInsert(KIO::UDSEntry::UDS_FILE_TYPE, S_IFDIR);
        entry.fastInsert(KIO::UDSEntry::UDS_NAME, device.m_names.at(0));
        entry.fastInsert(KIO::UDSEntry::UDS_ICON_NAME, "network-server");
        entry.fastInsert(KIO::UDSEntry::UDS_DISPLAY_TYPE, "DNS-SD Workstation");
        entry.fastInsert(KIO::UDSEntry::UDS_ICON_OVERLAY_NAMES, "/home/me/avahi.png");

//        entry.fastInsert(KIO::UDSEntry::UDS_URL, "discovery-view://dnssd/local/123");

        q->listEntries({ entry });
        entry.clear();
    }

    return std::nullopt;
}

std::optional<Error> Internal::get( const QUrl & url )
{
    qDebug() << url;

    const QString protocol = url.host();
    QString resource = url.path();
    resource.remove(resource.indexOf('/'), 1);
    qDebug() << protocol << resource;

    if (protocol == "upnp") {

    } else if (protocol == "dnssd") {

    } else if (protocol == "wsdiscovery") {

    }

    return std::nullopt;
}

class Slave : public KIO::SlaveBase
{
public:
    explicit Slave(const QByteArray &protocol, const QByteArray &pool_socket, const QByteArray &app_socket)
        : KIO::SlaveBase(protocol, pool_socket, app_socket)
        , d(new Internal(this))
    {
    }

    ~Slave() override = default;

    void get( const QUrl& url ) override
    {
        qDebug() << url;

        const auto opt = d->get(url);
        conclude(opt);
    }

    void stat(const QUrl& url) override
    {
        qDebug() << url;

        const auto opt = d->stat(url);
        conclude(opt);
    }

    void listDir(const QUrl& url) override
    {
        qDebug() << url;

        const auto opt = d->listDir(url);
        conclude(opt);
    }

private:
    // Concludes a command in either error or finish.
    void conclude(const std::optional<Error> &opt)
    {
        if (opt.has_value()) {
            error(opt);
        } else {
            finished();
        }
    }

    // Convenience
    void error(const std::optional<Error> &opt)
    {
        const auto e = opt.value();
        KIO::SlaveBase::error(e.code, e.msg);
    }

    Internal *d = nullptr;
};

extern "C" {
    Q_DECL_EXPORT int kdemain( int argc, char **argv )
    {
        QCoreApplication app(argc, argv);
        app.setApplicationName("kio_settings");
        qDebug() << "kdemain for settings kioslave";
        Slave slave(argv[1], argv[2], argv[3]);
        slave.dispatchLoop();
        return 0;
    }
}

